//-----------------------------------------------------------------------------
// File: Explosion.cpp
//
// Desc: Example code showing how to do volumetric explosions.
//       Note: This code uses the D3D Framework helper library.
//
//
// Copyright (c) 1999 NVIDIA Corporation. All rights reserved.
//-----------------------------------------------------------------------------


#include "Explosion.h"
#include "nvidia.h"
#include <d3dframework.h>

extern CD3DFramework* g_pFramework;
extern float g_fZBias;
extern BOOL g_bRenderBB;


#define MAXINDICES 4096
#define VERTEXBUFFERSIZE 1024
typedef struct tagTXPOINT{
	float x,y,z,rhw;
}TXPOINT, *PTXPOINT;

typedef struct tagF_RECT{
	float top,bottom,left,right;
}F_RECT, *PF_RECT;

typedef struct _PLANE{
	float A,B,C,D;
} PLANE, *LPPLANE;


//Some useful macros...
#define SQR(X) ((X)*(X))
#define CLAMP(X,L,H) ((X)<(L)?(L):((X)>(H)?(H):(X)))
#define LERP(m,x0,x1)       ((x0) + (m)*((x1)-(x0)))
#define INRANGE(v,min,max) ((v>=min && v<=max)?TRUE:FALSE)

//Clip codes
#define CLIP_TOP	0x01
#define CLIP_BOTTOM	0x02
#define CLIP_LEFT	0x04
#define CLIP_RIGHT	0x08
#define CLIP_FRONT  0x10
#define CLIP_BACK   0x20



//-----------------------------------------------------------------------------
// Name: VectorMatrixMultiply()
// Desc: Multiplies a vector by a matrix (assumes column 4 of matix is (0,0,0,1)
//-----------------------------------------------------------------------------
HRESULT VectorMatrixMultiply( D3DVECTOR& vDest, D3DVECTOR& vSrc,
                                      LPD3DMATRIX mat)
{
    vDest.x = vSrc.x*mat->_11 + vSrc.y*mat->_21 + vSrc.z* mat->_31 + mat->_41;
    vDest.y = vSrc.x*mat->_12 + vSrc.y*mat->_22 + vSrc.z* mat->_32 + mat->_42;
    vDest.z = vSrc.x*mat->_13 + vSrc.y*mat->_23 + vSrc.z* mat->_33 + mat->_43;

    return S_OK;
}
//-----------------------------------------------------------------------------
// Name: GetPositionFromViewMatrix()
// Desc: Calculates the view position (in world coordinates) given a view matrix
//-----------------------------------------------------------------------------
void GetPositionFromViewMatrix(D3DVECTOR &pos,D3DMATRIX *pView)
{
	pos.x=-(pView->_41*pView->_11 + pView->_42*pView->_12 + pView->_43*pView->_13);
	pos.y=-(pView->_41*pView->_21 + pView->_42*pView->_22 + pView->_43*pView->_23);
	pos.z=-(pView->_41*pView->_31 + pView->_42*pView->_32 + pView->_43*pView->_33);
}

//*************************************************************************
// Function name	: Contained
// Description	    : Utility function to check if any
//						part of a triangle is visible within a 2D 
//						rectangle. rectangle is constructed as the
//						bounding box of circle (x-cx)^2 + (y-cy)^2 = rad^2
//
// Return type		: BOOL 
//*************************************************************************
BOOL Contained(float cx,float cy, float rad, LPD3DTLVERTEX v1,LPD3DTLVERTEX v2,LPD3DTLVERTEX v3)
{
	unsigned char v1Out=0,v2Out=0,v3Out=0;
	BOOL inside=FALSE;	//used for stage 3

	F_RECT r={cy-rad,cy+rad,cx-rad,cx+rad};

	//Stage 1 - simple clip outcode check
	//v1
	if (v1->sy <= r.top) v1Out|= CLIP_TOP;
	if (v1->sy >= r.bottom) v1Out|= CLIP_BOTTOM;
	if (v1->sx <= r.left) v1Out|= CLIP_LEFT;
	if (v1->sx >= r.right) v1Out|= CLIP_RIGHT;
	
	if (0==v1Out) return TRUE; //Vertex is in rectangle

	//v2
	if (v2->sy <= r.top) v2Out|= CLIP_TOP;
	if (v2->sy >= r.bottom) v2Out|= CLIP_BOTTOM;
	if (v2->sx <= r.left) v2Out|= CLIP_LEFT;
	if (v2->sx >= r.right) v2Out|= CLIP_RIGHT;
	
	if (0==v2Out) return TRUE; //Vertex is in rectangle

	//v1
	if (v3->sy <= r.top) v3Out|= CLIP_TOP;
	if (v3->sy >= r.bottom) v3Out|= CLIP_BOTTOM;
	if (v3->sx <= r.left) v3Out|= CLIP_LEFT;
	if (v3->sx >= r.right) v3Out|= CLIP_RIGHT;
	
	if (0==v3Out) return TRUE; //Vertex is in rectangle

	if (v1Out & v2Out & v3Out) return FALSE; //all verts are outside same edge of rect

	//stage2
	float dx=0,dy=0,m=0,c=0,y=0;

	//find intersections of poly edges with bounding box
	//also check if poly wholly contains the bbox

	// edge v1v2
	dx = v2->sx-v1->sx; dy = v2->sy-v1->sy;
	m = dy/dx; c= v1->sy - m*v1->sx;

	y = m*r.left + c;
	if ( y >= r.bottom && y<=r.top)
		return TRUE; //edge crosses bbox

	y = m*r.right + c;
	if ( y <= r.bottom && y>=r.top)
		return TRUE; //edge crosses bbox

	if (y>=r.top)
		inside=!inside; //containment check

	//edge v2v3
	dx = v3->sx-v2->sx; dy = v3->sy-v2->sy;
	m = dy/dx; c= v2->sy - m*v2->sx;

	y = m*r.left + c;
	if ( y >= r.bottom && y<=r.top)
		return TRUE;

	y = m*r.right + c;
	if ( y <= r.bottom && y>=r.top)
		return TRUE;

	if (y>=r.top) inside=!inside;

	// edge v3v1
	dx = v1->sx-v3->sx; dy = v1->sy-v3->sy;
	m = dy/dx; c= v3->sy - m*v3->sx;

	y = m*r.left + c;
	if ( y >= r.bottom && y<=r.top)
		return TRUE;

	y = m*r.right + c;
	if ( y <= r.bottom && y>=r.top)
		return TRUE;

	if (y>=r.top) inside=!inside;

	return inside;


}

//*************************************************************************
// Function name	: SetTLVertex
// Description	    : Utility function to set fields of a D3D TL vertex
//					  Colour defaults to white, specular to 0.
// Return type		: void 
//*************************************************************************
void SetTLVertex(D3DTLVERTEX &v,float x,float y,float z, float rhw,float tu,float tv){
	v.sx = x;
	v.sy = y;
	v.sz = z;
	v.rhw =rhw;
	v.color = 0xFFFFFFFF;
	v.specular = 0;
	v.tu=tu;
	v.tv=tv;
}
//*************************************************************************
// Function name	: SetLVertex
// Description	    : Utility function to set fields of a D3D L vertex
//					  Colour defaults to white, specular to 0.
// Return type		: void 
//*************************************************************************
void SetLVertex(D3DLVERTEX &v,float x,float y,float z,float tu,float tv, DWORD col=0xFFFFFFFF){
	v.x = x;
	v.y = y;
	v.z = z;
	v.dwReserved=0;
	v.color = col;
	v.specular = 0;
	v.tu=tu;
	v.tv=tv;
}


//*************************************************************************
// Function name	: EdgePlaneIntersection
// Description	    : Compute the point of intersection of the vector from
//                    'start' to 'end' with the 3D plane 'plane'
// Return type		: BOOL - intersection valid
//
// Remarks          : Intersection is placed in 'out'. Calculation is performed
//                    such that result is the same regardless of ordering of
//                    endpoints.
//*************************************************************************
BOOL EdgePlaneIntersection(D3DVECTOR &out,PLANE *plane,D3DVECTOR &start,D3DVECTOR &end)
{
	if (end.x>start.x){
		D3DVECTOR delta(end.x-start.x,end.y-start.y,end.z-start.z);
		float denom=delta.x*plane->A + delta.y*plane->B + delta.z*plane->C;
		float num=start.x*plane->A + start.y*plane->B + start.z*plane->C-plane->D;

		if (denom==0.0f)
			return FALSE;
		float t=-num/denom;

		if (t<-0.01f||t>1.01f)
		{
			return FALSE;
		}
		out.x=start.x+t*delta.x;
		out.y=start.y+t*delta.y;
		out.z=start.z+t*delta.z;
		return TRUE;
	}
	else {
		D3DVECTOR delta(start.x-end.x,start.y-end.y,start.z-end.z);
		float denom=delta.x*plane->A + delta.y*plane->B + delta.z*plane->C;
		float num=end.x*plane->A + end.y*plane->B + end.z*plane->C-plane->D;

		if (denom==0.0f)
			return FALSE;
		float t=-num/denom;

		if (t<-0.01f||t>1.01f)
		{
			return FALSE;
		}
		out.x=end.x+t*delta.x;
		out.y=end.y+t*delta.y;
		out.z=end.z+t*delta.z;
		return TRUE;
	}

}
// static variables to hold details of explosion frustum
static PLANE Ptop;
static PLANE Pbottom;
static PLANE Pleft;
static PLANE Pright;
static PLANE Pfront;
static PLANE Pback;

//create some references to the A, B and C components of the planes as vectors for convenience.
D3DVECTOR &Nleft=*(D3DVECTOR *)&Pleft;
D3DVECTOR &Ntop=*(D3DVECTOR *)&Ptop;
D3DVECTOR &Nright=*(D3DVECTOR *)&Pright;
D3DVECTOR &Nbottom=*(D3DVECTOR *)&Pbottom;
D3DVECTOR &Nback=*(D3DVECTOR *)&Pback;
D3DVECTOR &Nfront=*(D3DVECTOR *)&Pfront;

//Used to hold the position of the front of the explosion in view coordinates
D3DVECTOR fview;



//*************************************************************************
// Function name	: CalculateBoundingFrustum
// Description	    : Calculates a bounding frustum for the explosion in 
//                    world space. Explosion frustum is clipped to view frustum
// Return type		: BOOL 
//*************************************************************************
BOOL CalculateBoundingFrustum(D3DVECTOR &c,float rad,D3DMATRIX *pViewMat, LPDIRECT3DDEVICE3 pDev)
{
	if (rad<0.0001)
		return FALSE;
	
	D3DMATRIX proj;
	D3DVECTOR e;
	GetPositionFromViewMatrix(e,pViewMat);

	D3DVECTOR ec=c-e; //Vector from eyepoint to center of the explosion
	ec=Normalize(ec);//normalise the vector

	//Calculate the centerpoint of the explosion in viewspace
	D3DVECTOR cview;
	VectorMatrixMultiply(cview,c,pViewMat);
	
		
	//generate a point on the front of the explosion and transform into viewspace
	D3DVECTOR frontC = c - rad*ec;
	VectorMatrixMultiply(fview,frontC,pViewMat);
	
	//generate billboard vertices
	
	//	p1    p2
	//    -----  -
	//   |     | | dy
	//   |  c  | -
	//   |     | 
	//    -----  
	//  p4    p3
	//   |--|
	//    dx
	D3DVECTOR dx,dy;
	D3DVECTOR p1,p2,p3,p4;
	
	dx.x=pViewMat->_11*rad;
	dx.y=pViewMat->_21*rad;
	dx.z=pViewMat->_31*rad;

	dy.x=pViewMat->_12*rad;
	dy.y=pViewMat->_22*rad;
	dy.z=pViewMat->_32*rad;
	
	p1=c-dx+dy;
	p2=c+dx+dy;
	p3=c+dx-dy;
	p4=c-dx-dy;


	//We need to get some info from the projection matrix to reconstruct the view frustum
	pDev->GetTransform(D3DTRANSFORMSTATE_PROJECTION, &proj);

	//find near clip plane
	float fNearPlane=-(proj._33)/proj._43 ;

	//find half height and half width
	float scale=1.0f/((2.0f*proj._34)/fNearPlane);
	float hheight=0.5f/(proj._22*scale);
	float hwidth=0.5f/(proj._11*scale);

	
	//get some orientation vectors from the view matrix
	D3DVECTOR v(pViewMat->_13,pViewMat->_23,pViewMat->_33); //view direction (unit length)
	D3DVECTOR u(pViewMat->_12,pViewMat->_22,pViewMat->_32); //up direction (unit length)
	D3DVECTOR l(pViewMat->_11,pViewMat->_21,pViewMat->_31); //left (increasing x) direction (unit length)

	/***********************************************************************************/
	/*****The following code makes extensive use if D3D overloaded vector operators*****/
	/***********************************************************************************/

	//build view frustum clip planes...
	Nleft=fNearPlane*v-hwidth*l-(fNearPlane+hwidth/fNearPlane)*v;
	Nleft=Normalize(Nleft);
	Pleft.D=DotProduct(Nleft,e);

	Nright=fNearPlane*v+hwidth*l-(fNearPlane+hwidth/fNearPlane)*v;
	Nright=Normalize(Nright);
	Pright.D=DotProduct(Nright,e);

	Ntop=fNearPlane*v+hheight*u-(fNearPlane+hheight/fNearPlane)*v;
	Ntop=Normalize(Ntop);
	Ptop.D=DotProduct(Ntop,e);
	
	Nbottom=fNearPlane*v-hheight*u-(fNearPlane+hheight/fNearPlane)*v;
	Nbottom=Normalize(Nbottom);
	Pbottom.D=DotProduct(Nbottom,e);

	//clip billboard to view frustum (assumes billboard is axis aligned)
	if (EdgePlaneIntersection(p1,&Pleft,p2,p1)){
		EdgePlaneIntersection(p4,&Pleft,p4,p3);		
	}
	if (EdgePlaneIntersection(p2,&Pright,p2,p1)){
		EdgePlaneIntersection(p3,&Pright,p4,p3);
	}

	if (EdgePlaneIntersection(p1,&Ptop,p1,p4)){
		EdgePlaneIntersection(p2,&Ptop,p2,p3);
	}
	if (EdgePlaneIntersection(p4,&Pbottom,p1,p4)){
		EdgePlaneIntersection(p3,&Pbottom,p2,p3);
	}


	//build explosion frustum from eyepoint and clipped billboard
	D3DVECTOR ep1=e-p1;
	D3DVECTOR ep2=e-p2;
	D3DVECTOR ep3=e-p3;
	D3DVECTOR ep4=e-p4;


	//Generate Planes
	Nleft=CrossProduct(ep4,ep1);
	Nleft=Normalize(Nleft);
	Pleft.D=Nleft.x*e.x + Nleft.y*e.y + Nleft.z*e.z;
	

	Ntop=CrossProduct(ep1,ep2);
	Ntop=Normalize(Ntop);
	Ptop.D=Ntop.x*e.x + Ntop.y*e.y + Ntop.z*e.z;
	
	Nright=CrossProduct(ep2,ep3);
	Nright=Normalize(Nright);
	Pright.D=Nright.x*e.x + Nright.y*e.y + Nright.z*e.z;
	
	Nbottom=CrossProduct(ep3,ep4);
	Nbottom=Normalize(Nbottom);
	Pbottom.D=Nbottom.x*e.x + Nbottom.y*e.y + Nbottom.z*e.z;

	
	Pback.A=pViewMat->_13;
	Pback.B=pViewMat->_23;
	Pback.C=pViewMat->_33;
	Pback.D=Pback.A*c.x+Pback.B*c.y+Pback.C*c.z;

	Pfront.A=-pViewMat->_13;
	Pfront.B=-pViewMat->_23;
	Pfront.C=-pViewMat->_33;

	//decide if we need to use near clip plane instead of explosion front plane
	
	if ((frontC.x-e.x)*pViewMat->_13+(frontC.y-e.y)*pViewMat->_23+(frontC.z-e.z)*pViewMat->_33<0)
	{
		//front point behind eye
		Pfront.D=Pfront.A*(e.x+fNearPlane*pViewMat->_13)+Pfront.B*(e.y+fNearPlane*pViewMat->_23)+Pfront.C*(e.z+fNearPlane*pViewMat->_33);
	}
	else
	{
		Pfront.D=Pfront.A*frontC.x + Pfront.B*frontC.y + Pfront.C*frontC.z;
	}

	return TRUE;
}

inline BOOL EdgePlaneIntersection(D3DLVERTEX &out,PLANE *plane,D3DLVERTEX &start,D3DLVERTEX &end)
{
	return EdgePlaneIntersection(*(D3DVECTOR *)&out,plane,*(D3DVECTOR *)&start,*(D3DVECTOR *)&end);
}

//checks to see if pt is 'inside' plane
inline BOOL Inside(D3DLVERTEX *pt,PLANE *plane)
{
	return ((pt->x*plane->A + pt->y*plane->B + pt->z*plane->C) <= plane->D);
}


//*************************************************************************
// Function name	: ClipPolyToPlane
// Description	    : clip polygon inPoly to the plane, returning result in
//                    outPoly.
//                    (Adapted from Foley et al,1990)
// Return type		: void 
//*************************************************************************
void ClipPolyToPlane(PLANE *plane,D3DLVERTEX *inPoly,int inSize,D3DLVERTEX *outPoly,int *outSize){

	int i,s,e;
	D3DLVERTEX temp;
	*outSize=0;
	for (i=0;i<inSize;i++)
	{
		s=i;e=(i+1)%inSize;

		if (Inside(&inPoly[e],plane))
		{
			if (Inside(&inPoly[s],plane))
			{
				outPoly->x=inPoly[e].x;
				outPoly->y=inPoly[e].y;
				outPoly->z=inPoly[e].z;
				(*outSize)++;outPoly++;
			}
			else
			{
				EdgePlaneIntersection(*outPoly,plane,inPoly[s],inPoly[e]);
				(*outSize)++;outPoly++;
				outPoly->x=inPoly[e].x;
				outPoly->y=inPoly[e].y;
				outPoly->z=inPoly[e].z;
				(*outSize)++;outPoly++;
			}
		}
		else
		{
			if (Inside(&inPoly[s],plane))
			{
				EdgePlaneIntersection(*outPoly,plane,inPoly[s],inPoly[e]);
				(*outSize)++;outPoly++;
			}
		}
	}

}



//*************************************************************************
// Function name	: CalculateUVCoordinates
// Description	    : Generate planar projected UV coordinates for each of
//                    the vertices in pVerts. Also calculates the intensity
//                    at each of the points. Assumes vertices are presented
//                    in world coordinates.
// Return type		: void 
//*************************************************************************
void CalculateUVCoordinates(D3DMATRIX *pViewMat,D3DVECTOR &cview,float rad,D3DLVERTEX *pVerts,DWORD count)
{
	D3DVECTOR vview;
	float intensity,uvscalefactor;
	BYTE col;
	
	
	//Generate uv coordinates and intensities
	for(int i=0;i<count;i++)
	{
		//transform vertex into view space
		VectorMatrixMultiply(vview,*(D3DVECTOR *)&pVerts[i],pViewMat);
		
		//generate intensity
		//linear interpolation between front of explosion in view space,(fview) and center (cview)
		intensity=255.0*(vview.z-fview.z)/(cview.z-fview.z);
		col=CLAMP(intensity,0.0f,255.0f);
		pVerts[i].color=RGBA_MAKE(col,col,col,255);
		pVerts[i].specular=0;
		
		//generate uv coordinates
		//commented block is less optimal but easier to understand!
/*		float cx,cy;
		cx=cview.x*vview.z/cview.z;
		cy=cview.y*vview.z/cview.z;
		uvscalefactor=cview.z/(rad*2.0f*vview.z);

		pVerts[i].tu=0.5f+(vview.x-cx)*uvscalefactor;
		pVerts[i].tv=0.5f+(vview.y-cy)*uvscalefactor;
*/
		uvscalefactor=1.0f/(rad*2.0f*vview.z);
		pVerts[i].tu=0.5f+(vview.x*cview.z-vview.z*cview.x)*uvscalefactor;
		pVerts[i].tv=0.5f-(vview.y*cview.z-vview.z*cview.y)*uvscalefactor;

	}

}



//*************************************************************************
// Function name	: PointPlaneCoeffDW
// Description	    : Helper function which gives the dot product of
//                    a point and a plane normal minus the plane 'offset' as
//                    an signed number - exact value is irrelevant as this is
//                    only used as an inside/outside test - quicker than doing
//                    a floating point compare.
// Return type		: int 
//*************************************************************************
inline int PointPlaneCoeffDW(const D3DVECTOR *pt,PLANE *p)
{
	float val=pt->x*p->A + pt->y*p->B + pt->z*p->C - p->D; 
	return *(int *)&val;
}

float DotProduct(const D3DVECTOR *v1,const D3DVECTOR *v2)
{
	return v1->x*v2->x + v1->y*v2->y + v1->z*v2->z;
}

//*************************************************************************
// Function name	: Clip
// Description	    : Clip the triangle defined by v1v2v3 to the frustum
//					  defined by CalculateBoundingFrustum
//
// Return type		: BOOL 
//*************************************************************************
BOOL Clip(D3DVECTOR &cview,float rad, D3DVECTOR *v1,D3DVECTOR *v2,D3DVECTOR *v3,D3DMATRIX *pViewMat,D3DLVERTEX *pOut,int *pNum)
{
	unsigned char v1Out=0,v2Out=0,v3Out=0;

	*pNum=0;
	//Stage 1 - simple clip outcode check
	
	if (PointPlaneCoeffDW(v1,&Pback)   >= 0)   v1Out|= CLIP_BACK;
	if (PointPlaneCoeffDW(v2,&Pback)   >= 0)   v2Out|= CLIP_BACK;
	if (PointPlaneCoeffDW(v3,&Pback)   >= 0)   v3Out|= CLIP_BACK;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}
	if (!(v1Out&CLIP_BACK) && PointPlaneCoeffDW(v1,&Pfront)  >= 0)  v1Out|= CLIP_FRONT;
	if (!(v2Out&CLIP_BACK) && PointPlaneCoeffDW(v2,&Pfront)  >= 0)  v2Out|= CLIP_FRONT;
	if (!(v3Out&CLIP_BACK) && PointPlaneCoeffDW(v3,&Pfront)  >= 0)  v3Out|= CLIP_FRONT;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}

	if (PointPlaneCoeffDW(v1,&Pleft)   >= 0)   v1Out|= CLIP_LEFT;
	if (PointPlaneCoeffDW(v2,&Pleft)   >= 0)   v2Out|= CLIP_LEFT;
	if (PointPlaneCoeffDW(v3,&Pleft)   >= 0)   v3Out|= CLIP_LEFT;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}
	if (!(v1Out&CLIP_LEFT) && PointPlaneCoeffDW(v1,&Pright)  >= 0)  v1Out|= CLIP_RIGHT;
	if (!(v2Out&CLIP_LEFT) && PointPlaneCoeffDW(v2,&Pright)  >= 0)  v2Out|= CLIP_RIGHT;
	if (!(v3Out&CLIP_LEFT) && PointPlaneCoeffDW(v3,&Pright)  >= 0)  v3Out|= CLIP_RIGHT;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}
	if (PointPlaneCoeffDW(v1,&Ptop)    >= 0)    v1Out|= CLIP_TOP;
	if (PointPlaneCoeffDW(v2,&Ptop)    >= 0)    v2Out|= CLIP_TOP;
	if (PointPlaneCoeffDW(v3,&Ptop)    >= 0)    v3Out|= CLIP_TOP;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}
	
	if (!(v1Out&CLIP_TOP) && PointPlaneCoeffDW(v1,&Pbottom) >= 0) v1Out|= CLIP_BOTTOM;
	if (!(v2Out&CLIP_TOP) && PointPlaneCoeffDW(v2,&Pbottom) >= 0) v2Out|= CLIP_BOTTOM;
	if (!(v3Out&CLIP_TOP) && PointPlaneCoeffDW(v3,&Pbottom) >= 0) v3Out|= CLIP_BOTTOM;
	if (v1Out & v2Out & v3Out) {
		//all verts are outside same edge of rect
		return FALSE; 
	}
	
	//clip the poly....
	pOut[0].x=v1->x;pOut[0].y=v1->y;pOut[0].z=v1->z;
	pOut[1].x=v2->x;pOut[1].y=v2->y;pOut[1].z=v2->z;
	pOut[2].x=v3->x;pOut[2].y=v3->y;pOut[2].z=v3->z;
	(*pNum)=3;

	if (!(0==v1Out && 0==v2Out && 0==v3Out))
	{
		D3DLVERTEX temp[16];
		int tCount=0;
		ClipPolyToPlane(&Ptop,pOut,*pNum,temp,&tCount);
		ClipPolyToPlane(&Pbottom,temp,tCount,pOut,pNum);
		ClipPolyToPlane(&Pfront,pOut,*pNum,temp,&tCount);
		ClipPolyToPlane(&Pback,temp,tCount,pOut,pNum);
		ClipPolyToPlane(&Pleft,pOut,*pNum,temp,&tCount);		
		ClipPolyToPlane(&Pright,temp,tCount,pOut,pNum);
		
		

	}

	//generate UV coordinates and intensities
	CalculateUVCoordinates(pViewMat,cview,rad,pOut,(*pNum));
	return (*pNum>0);
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Explosion::Explosion()
{

	//set initial position and size of explosion
	x=0;
	z=-60.f;
	y=0.0f;
	radius=15.;
	
	//increment used to animate explosion
	inc=0.5f;

	//Create some vertex buffers
	//pVB and pVBTX are used in screenspace version only
	//pVBExp is used in worldspace version
	D3DVERTEXBUFFERDESC vbDesc;

	pVB=NULL;pVBTX=NULL;pVBExp=NULL;
	memset( &vbDesc,0, sizeof(D3DVERTEXBUFFERDESC) );
	vbDesc.dwSize        = sizeof(D3DVERTEXBUFFERDESC);
	vbDesc.dwCaps        = 0L;
	vbDesc.dwFVF         = D3DFVF_XYZ ;
	vbDesc.dwNumVertices = 4;

	//First vertex buffer to hold untransformed sphere coordinates
	g_pFramework->GetDirect3D()->CreateVertexBuffer(&vbDesc,&pVB,0,NULL);

	//second vertex buffer to hold transformed sphere coordinates
	vbDesc.dwFVF		= D3DFVF_XYZRHW;
	g_pFramework->GetDirect3D()->CreateVertexBuffer(&vbDesc,&pVBTX,0,NULL);

	//3rd vertex buffer to hold explosion
	vbDesc.dwCaps		= D3DVBCAPS_WRITEONLY;
	vbDesc.dwFVF		= D3DFVF_LVERTEX;
	vbDesc.dwNumVertices= VERTEXBUFFERSIZE;
	g_pFramework->GetDirect3D()->CreateVertexBuffer(&vbDesc,&pVBExp,0,NULL);
}

Explosion::~Explosion()
{
	//Clean up
	if (pVB)
		pVB->Release();
	if (pVBTX)
		pVBTX->Release();
	if (pVBExp)
		pVBExp->Release();
}

void Explosion::IncrementRadius(void)
{
	///////////////////////////////////////////////////
	//	Gradually change explosion radius
	///////////////////////////////////////////////////
	radius+=inc;
	if (radius>=32.0f) radius=0.5f;
}

void Explosion::RenderZBias(LPDIRECT3DDEVICE3 pDev, MyLandscape *pLandscape,LPD3DMATRIX pViewMat)
{
	
	D3DVECTOR centre(x,y,z);
	D3DVECTOR e;
	
	D3DLVERTEX backQuad[4];
	D3DVECTOR cview;
	float u,v;

	GetPositionFromViewMatrix(e,pViewMat);
	
	D3DMATRIX vmatInvB,tmat;
	vmatInvB._11=pViewMat->_11; vmatInvB._12=pViewMat->_21; vmatInvB._13=pViewMat->_31; vmatInvB._14=0;
	vmatInvB._21=pViewMat->_12; vmatInvB._22=pViewMat->_22; vmatInvB._23=pViewMat->_32; vmatInvB._24=0;
	vmatInvB._31=pViewMat->_13; vmatInvB._32=pViewMat->_23; vmatInvB._33=pViewMat->_33; vmatInvB._34=0;
	vmatInvB._41=x;vmatInvB._42=y;vmatInvB._43=z;vmatInvB._44=1;
	
	
	D3DVECTOR delta(x-e.x,y-e.y,z-e.z);
	float dis=Magnitude(delta);
	
	VectorMatrixMultiply(cview,centre,pViewMat);

	if (!CalculateBoundingFrustum(centre,radius,pViewMat,pDev))
		return;
	
	
	SetLVertex(backQuad[3],-radius,-radius,0,0.0f,1.0f);
	SetLVertex(backQuad[2],+radius,-radius,0,1.0f,1.0f);
	SetLVertex(backQuad[1],+radius,+radius,0,1.0f,0.0f);
	SetLVertex(backQuad[0],-radius,+radius,0,0.0f,0.0f);

	

	//-----------------------------------------------
	//Set renderstates for landscape pass
	//-----------------------------------------------

	D3DMATRIX matProj;

	//Set projection matrix such that g_fZBias is added to each of the post-transform z coordinates
	D3DUtil_SetProjectionMatrix2(matProj, 90.0f, 1.0f, 1000.0f,g_fZBias);
	pDev->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &matProj );	

	pDev->SetRenderState( D3DRENDERSTATE_ZFUNC,D3DCMP_LESSEQUAL);

	pDev->SetRenderState( D3DRENDERSTATE_SRCBLEND,D3DBLEND_ONE);
	pDev->SetRenderState( D3DRENDERSTATE_DESTBLEND,D3DBLEND_ONE);

	D3DLVERTEX *exp;
	pVBExp->Lock( DDLOCK_WRITEONLY|DDLOCK_SURFACEMEMORYPTR |DDLOCK_NOSYSLOCK,(LPVOID *)&exp,NULL);
	
	D3DVECTOR *v1,*v2,*v3;
	int numTris=pLandscape->x*pLandscape->z*2;
	D3DLVERTEX pVerts[16];
	int num=0;
	int k;
	int count=0;
	WORD indices[MAXINDICES];
	int totalverts=0;

	for (int l=0;l<numTris;l++){

		//copy vertex refs of the triangle to local variables 

		v1=(D3DVECTOR *)&pLandscape->pVertices[pLandscape->pTriangles[l].v1];
		v2=(D3DVECTOR *)&pLandscape->pVertices[pLandscape->pTriangles[l].v2];
		v3=(D3DVECTOR *)&pLandscape->pVertices[pLandscape->pTriangles[l].v3];
		
		//clip this triangle to the view frustum and re render if necessary...
		if (Clip(cview,radius,v1,v2,v3,pViewMat,pVerts,&num))
		{

			//copy clipped tris for rendering later...
			//we need to triangulate the poly
			//we use a fan rooted at the most intense vertex to avoid starnge artifacts

			//find highest intensity vert ( and copy verts into vertex buffer)
			int intensity=pVerts[0].color&0xff;
			int ind=0,maxin=intensity;
#ifdef _DEBUG
				if (totalverts>VERTEXBUFFERSIZE)
				{
					OutputDebugString("Vertex buffer overflow in RenderZBias\n");
					break;
				}
#endif
			(*exp)=pVerts[0];
			exp++;
			totalverts++;
			for (k=1;k<num;k++)
			{
				intensity=pVerts[k].color&0xff;
				if (intensity>maxin)
				{
					maxin=intensity;
					ind=k;
				}
				(*exp)=pVerts[k];
				exp++;
				totalverts++;
#ifdef _DEBUG
				if (totalverts>VERTEXBUFFERSIZE)
				{
					OutputDebugString("Vertex buffer overflow in RenderZBias\n");
					break;
				}
#endif

			}
			//triangulate the fan into the index list
			int basecount=totalverts-num;
			for (k=0;k<num-2;k++)
			{
#ifdef _DEBUG
				if (count>MAXINDICES-3)
				{
					OutputDebugString("Index array overflow in RenderZBias\n");
					break;
				}
#endif
				indices[count++]=ind+basecount;
				indices[count++]=((ind+k+1)%num) + basecount;
				indices[count++]=((ind+k+2)%num) + basecount;
			}

		}
	}

	pVBExp->Unlock();
	
	//Draw the explosion pass
	pDev->DrawIndexedPrimitiveVB(D3DPT_TRIANGLELIST,pVBExp,indices,count,D3DDP_DONOTLIGHT|D3DDP_DONOTUPDATEEXTENTS|D3DDP_DONOTCLIP);

	
	//-----------------------------------------------
	//Render back plane to frame buffer (z test < to avoid double writes)
	//-----------------------------------------------
	pDev->SetRenderState( D3DRENDERSTATE_ZFUNC,D3DCMP_LESS);
	pDev->SetRenderState( D3DRENDERSTATE_ZWRITEENABLE,TRUE);
	
	pDev->GetTransform(D3DTRANSFORMSTATE_WORLD,&tmat);
	pDev->SetTransform(D3DTRANSFORMSTATE_WORLD,&vmatInvB);
	if (g_bRenderBB)
		pDev->DrawPrimitive(D3DPT_TRIANGLEFAN,D3DFVF_LVERTEX,(LPVOID)backQuad,4,D3DDP_DONOTLIGHT |D3DDP_DONOTUPDATEEXTENTS);
	pDev->SetTransform(D3DTRANSFORMSTATE_WORLD,&tmat);
	

	D3DUtil_SetProjectionMatrix2(matProj, 90.0f, 1.0f, 1000.0f,0.0f );
	pDev->SetTransform( D3DTRANSFORMSTATE_PROJECTION, &matProj );

	
}



//*************************************************************************
// Function name	: Render
// Description	    : Render the explosion over the current landscape.
// Return type		: void 
//*************************************************************************
void Explosion::Render(LPDIRECT3DDEVICE3 pDev, MyLandscape *pLandscape,LPD3DMATRIX pViewMat)
{
	

	D3DTLVERTEX tlv1,tlv2,tlv3,tlv4;
	LPD3DTLVERTEX v1,v2,v3;
	
	///////////////////////////////////////////////////
	// Add centre of sphere and a point radius units away
	// along the screen x axis to a vertex buffer
	///////////////////////////////////////////////////
	LPD3DVECTOR pvbVerts=NULL;
	PTXPOINT pTx=NULL;

	pVB->Lock( DDLOCK_WRITEONLY|DDLOCK_SURFACEMEMORYPTR |DDLOCK_NOSYSLOCK,(LPVOID *)&pvbVerts,NULL);

	pvbVerts[0].x=x;
	pvbVerts[0].y=y;
	pvbVerts[0].z=z;

	//build a point to side of explosion
	pvbVerts[1].x=x+radius*pViewMat->_11;
	pvbVerts[1].y=y+radius*pViewMat->_21;
	pvbVerts[1].z=z+radius*pViewMat->_31;

	//build a point to front of explosion
	pvbVerts[2].x=x-radius*pViewMat->_13;
	pvbVerts[2].y=y-radius*pViewMat->_23;
	pvbVerts[2].z=z-radius*pViewMat->_33;
	
	//build a point to rear of explosion
	pvbVerts[3].x=x+radius*pViewMat->_13;
	pvbVerts[3].y=y+radius*pViewMat->_23;
	pvbVerts[3].z=z+radius*pViewMat->_33;
	
	pVB->Unlock();

	//Transform the verts into screen space...
	pVBTX->ProcessVertices(D3DVOP_TRANSFORM,0,4,pVB,0,pDev,0);
	
	pVBTX->Lock( DDLOCK_READONLY|DDLOCK_SURFACEMEMORYPTR |DDLOCK_NOSYSLOCK,(LPVOID *)&pTx,NULL);


	//Read back the transformed screen coordinates
	float sx = pTx[0].x;
	float sy = pTx[0].y;
	float sz = pTx[0].z;
	float rhw = pTx[0].rhw;

	//Radius (in screen space) is the difference between the 2 x-coordinates of the first 2 points

	float r  = pTx[1].x-sx;
	float rtimes2 = r*2.0f;

	float frontz = pTx[2].z;

	float backz = pTx[3].z;
	float backrhw = pTx[3].rhw;

	//useful constants for inner loop
	float oortimes2=1.0f/rtimes2;
	float oozrange=1.0f/(sz-frontz);


	pVBTX->Unlock();
	
	//render the explosion


	//draw billboard to rear of explosion - reduces number of tris
	//to be drawn in second pass
	SetTLVertex(tlv1,sx-r,sy-r,backz,backrhw,0.0f,0.0f);
	SetTLVertex(tlv2,sx+r,sy-r,backz,backrhw,1.0f,0.0f);
	SetTLVertex(tlv3,sx+r,sy+r,backz,backrhw,1.0f,1.0f);
	SetTLVertex(tlv4,sx-r,sy+r,backz,backrhw,0.0f,1.0f);
	
	//Set z compares to strictly less than. Avoids artifacts on boundary
	//between billboard and multipass triangles
	pDev->SetRenderState(D3DRENDERSTATE_ZFUNC,D3DCMP_LESS);	
	if (g_bRenderBB){	
		pDev->Begin(D3DPT_TRIANGLELIST,D3DFVF_TLVERTEX,D3DDP_DONOTLIGHT|D3DDP_DONOTUPDATEEXTENTS);
		pDev->Vertex((void *)&tlv1);
		pDev->Vertex((void *)&tlv2);
		pDev->Vertex((void *)&tlv3);

		pDev->Vertex((void *)&tlv3);
		pDev->Vertex((void *)&tlv4);
		pDev->Vertex((void *)&tlv1);

		pDev->End(0);
	}


	//Now set z compares to exactly equal for second pass
	pDev->SetRenderState(D3DRENDERSTATE_ZFUNC,D3DCMP_EQUAL);
	
	//Turn off z writes to save bandwidth on the card
	pDev->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,0);

	
	//Now we need to run through all of the screenspace landscape tris
	//and re-render those which intersect the explosion volume.
	LPD3DTLVERTEX pTLVerts=NULL;
	 
	//used during intesity calcs
	float t1;
	BYTE c=0;
	
	//total number of triangles in the landscape
	int numTris=pLandscape->x*pLandscape->z*2;
	

	//Get access to the screenspace vertices already rendered
	pLandscape->pVBScreenSpace->Lock( DDLOCK_SURFACEMEMORYPTR |DDLOCK_NOSYSLOCK,(LPVOID *)&pTLVerts,NULL);
	
	
	//Use Begin/end for rendering to save me from buffering the vertices
	pDev->Begin(D3DPT_TRIANGLELIST,D3DFVF_TLVERTEX,D3DDP_DONOTLIGHT|D3DDP_DONOTUPDATEEXTENTS);
	

	for (int l=0;l<numTris;l++){

		//copy vertex refs of the triangle to local variables 

		v1=&pTLVerts[pLandscape->pTriangles[l].v1];
		v2=&pTLVerts[pLandscape->pTriangles[l].v2];
		v3=&pTLVerts[pLandscape->pTriangles[l].v3];
		
		//check if any part of tri is in explosion area
		if (( v1->sz>=frontz || v2->sz>=frontz || v3->sz>=frontz) && //tri behind front of explosion
			 ( v1->sz<=backz || v2->sz<=backz  || v3->sz<=backz) && //tri in front of rear of explosion
			Contained(sx,sy,r,v1,v2,v3)) //2d extent check..
		{
			//generate texture coordinates for this triangle
			//(planar projection)
			v1->tu=0.5f+(v1->sx-sx)*oortimes2;
			v1->tv=0.5f+(v1->sy-sy)*oortimes2;

			v2->tu=0.5f+(v2->sx-sx)*oortimes2;
			v2->tv=0.5f+(v2->sy-sy)*oortimes2;

			v3->tu=0.5f+(v3->sx-sx)*oortimes2;
			v3->tv=0.5f+(v3->sy-sy)*oortimes2;


			//modulate vertex intensities dependant on distance from center of explosion
			//this does not take into account perspective foreshortening of screen z
			//NOTE: for the intensity to be truly effective we need colour components greater than 1.0 (or 255)
			// and less then 0. We should really clip to polys
			t1=255.0*(v1->sz-frontz)*oozrange;
			c=CLAMP(t1,0.0f,255.0f);
			v1->color=RGBA_MAKE(c,c,c,255);
			
			t1=255.0*(v2->sz-frontz)*oozrange;
			c=CLAMP(t1,0.0f,255.0f);
			v2->color=RGBA_MAKE(c,c,c,255);

			t1=255.0*(v3->sz-frontz)*oozrange;
			c=CLAMP(t1,0.0f,255.0f);
			v3->color=RGBA_MAKE(c,c,c,255);
			
			//render this triangle	

			pDev->Vertex((void *)v1);
			pDev->Vertex((void *)v2);
			pDev->Vertex((void *)v3);

		}
	}
	//end the batch
	pDev->End(0L);

	//don't forget to unlock the landscape vertex buffer since we need to use it next frame!
	pLandscape->pVBScreenSpace->Unlock();

	
}

void Explosion::RenderB(LPDIRECT3DDEVICE3 pDev, MyLandscape *pLandscape,LPD3DMATRIX pViewMat)
{

	
	D3DLVERTEX lv1,lv2,lv3,lv4;
	
	//render the explosion
	//uses the inverse of the view matrix to control the orientation of the billboard
	//the translation part is set to be the position of the billboard
	D3DMATRIX tmat,vmatInv;

	vmatInv._11=pViewMat->_11;vmatInv._12=pViewMat->_21;vmatInv._13=pViewMat->_31;vmatInv._14=0;
	vmatInv._21=pViewMat->_12;vmatInv._22=pViewMat->_22;vmatInv._23=pViewMat->_32;vmatInv._24=0;
	vmatInv._31=pViewMat->_13;vmatInv._32=pViewMat->_23;vmatInv._33=pViewMat->_33;vmatInv._34=0;
	vmatInv._41=x;vmatInv._42=y;vmatInv._43=z;vmatInv._44=1;

	pDev->GetTransform(D3DTRANSFORMSTATE_WORLD,&tmat);
	pDev->SetTransform(D3DTRANSFORMSTATE_WORLD,&vmatInv);
	
	SetLVertex(lv1,-radius,-radius,0,0.0f,1.0f);
	SetLVertex(lv2,+radius,-radius,0,1.0f,1.0f);
	SetLVertex(lv3,+radius,+radius,0,1.0f,0.0f);
	SetLVertex(lv4,-radius,+radius,0,0.0f,0.0f);


	pDev->SetRenderState(D3DRENDERSTATE_ZFUNC,D3DCMP_LESSEQUAL);
	pDev->Begin(D3DPT_TRIANGLELIST,D3DFVF_LVERTEX,D3DDP_DONOTLIGHT|D3DDP_DONOTUPDATEEXTENTS);
	pDev->Vertex((void *)&lv1);
	pDev->Vertex((void *)&lv3);
	pDev->Vertex((void *)&lv2);

	pDev->Vertex((void *)&lv3);
	pDev->Vertex((void *)&lv1);
	pDev->Vertex((void *)&lv4);

	pDev->End(0);
	pDev->SetTransform(D3DTRANSFORMSTATE_WORLD,&tmat);

}